from APMonitor.apm import *
import numpy as np
import threading
import time, random

class ThreadClass(threading.Thread):
    def __init__(self, id, server, app, a, b):
        self.id = id
        self.server = server
        self.app = app
        self.a = a
        self.b = b
        self.objective = float('NaN')

        threading.Thread.__init__(self)
    def run(self):

        # Don't overload server by executing all scripts at once
        sleep_time = random.random()
        time.sleep(sleep_time)

        print('Running application ' + self.app + '\n')

        # Clear previous application
        apm(self.server,self.app,'clear all')

        # Load model file
        apm_load(self.server,self.app,'nlp.apm')

        # Designate parameters to change
        apm_info(self.server,self.app,'FV','a')
        apm_info(self.server,self.app,'FV','b')

        # Change parameter values
        apm_meas(self.server,self.app,'a',self.a)
        apm_meas(self.server,self.app,'b',self.b)

        # Solve on APM server
        solver_output = apm(self.server,self.app,'solve')

        # Retrieve solution status
        status = apm_tag(self.server,self.app,'apm.appstatus')

        # Retrieve results if the result is good
        if (status==1):
            sol = apm_sol(self.server,self.app)
            objective = sol['ob']
            self.objective = objective
        else:
            self.objective = float('NaN')

# Select server
server = 'http://byu.apmonitor.com'

# Optimize at mesh points
x = np.arange(20.0, 30.0, 2.0)
y = np.arange(30.0, 50.0, 2.0)
a, b = np.meshgrid(x, y)

# Array of threads
threads = []

# Calculate objective at all meshgrid points

# Load applications
id = -1
for i in range(0,a.shape[0]):
    for j in range(0,b.shape[1]):
        # Label threads with ID
        id = id + 1
        # Application name
        app = 'nlp_' + str(id) 
        # Create new thread
        threads.append(ThreadClass(id, server, app, a[i,j], b[i,j]))
        
# Run applications simultaneously as multiple threads
# Max number of threads to run at once
max_threads = 8
for t in threads:
    while (threading.activeCount()>max_threads):
        # check for additional threads every 0.01 sec
        time.sleep(0.01)
    # start the thread
    t.start()

# Check for completion
mt = 3.0 # max time
it = 0.0 # incrementing time
st = 1.0 # sleep time
while (threading.activeCount()>=1):
    time.sleep(st)
    it = it + st
    print('Active Threads: ' + str(threading.activeCount()))
    # Terminate after max time
    if (it>=mt):
        break

# Wait for all threads to complete
#for t in threads:
#    t.join()
#print('Threads complete')

# Initialize array for objective
obj = np.empty_like(a)

# Retrieve objective results
id = -1
for i in range(0,a.shape[0]):
    for j in range(0,b.shape[1]):
        id = id + 1
        obj[i,j] = threads[id].objective

# plot 3D figure of results
from mpl_toolkits.mplot3d import Axes3D
import matplotlib.pyplot as plt
from matplotlib import cm
import numpy as np

fig = plt.figure()
ax = fig.gca(projection='3d')
surf = ax.plot_surface(a, b, obj, \
                       rstride=1, cstride=1, cmap=cm.coolwarm, \
                       vmin = 12, vmax = 22, linewidth=0, antialiased=False)
ax.set_xlabel('a')
ax.set_ylabel('b')
ax.set_zlabel('obj')
ax.set_title('Multi-Threaded APM Python')
plt.show()
